/* LOADGIF */

/* mehrere Teilbilder werden zusammen in Gesamtbild gezeichnet */
/* kommt mit globaler oder lokaler Farbpalette zurecht, aber */
/* nicht mit unterschiedlichen, aspect ratio wird ignoriert */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "dib.h"

#pragma pack(1)

#define MAX_CODES 4095
static int curr_size;            /* The current code size */
static int clear;                /* Value for a clear code */
static int ending;               /* Value for a ending code */
static int newcodes;             /* First available code */
static int top_slot;             /* Highest code for current size */
static int slot;                 /* Last read code */
static int nbits_left;           /* # bits left in current byte */
static BYTE *bufptr;             /* Pointer to next byte in block */
static BYTE *bufend;
static BYTE dstack[MAX_CODES+1]; /* Stack for storing pixels */
static BYTE suffix[MAX_CODES+1]; /* Suffix table */
static WORD prefix[MAX_CODES+1]; /* Prefix linked list */
static BYTE byte_buff[256];

static int get_next_code(FILE *f)
{
    static int mask[13]={0,1,3,7,0xF,0x1F,0x3F,0x7F,0xFF,0x1FF,0x3FF,0x7FF,0xFFF};
    static BYTE b;
    int i,retval;

    retval=b>>(8-nbits_left);
    while(curr_size>nbits_left)
    {
        if(bufptr>=bufend)
        {
            i=getc(f);
            if(i<0) return -1;
            if(i) if((int)fread(byte_buff,1,i,f)!=i) return -1;
            bufptr=byte_buff;
            bufend=bufptr+i;
        }
        b=*bufptr++;
        retval|=(unsigned int)b<<nbits_left;
        nbits_left+=8;
    }
    nbits_left-=curr_size;
    return retval&mask[curr_size];
}

int decoder(FILE *f,unsigned int left,unsigned int top,unsigned int width,unsigned int height,char huge *data,long skew,unsigned int fullwidth,unsigned int fullheight,int interlaced)
{
    BYTE *sp;
    BYTE huge *ptr;
    int code,fc,oc,c,size;
    unsigned int x,y;

    if((size=getc(f))<2||size>9) return -1;
    curr_size=size+1;
    top_slot=1<<curr_size;
    clear=1<<size;
    ending=clear+1;
    slot=newcodes=ending+1;
    nbits_left=0;
    bufptr=bufend=byte_buff;
    x=y=oc=fc=0;
    sp=dstack;
    ptr=data+skew*(fullheight-1-top-y)+left;
    while((c=get_next_code(f))!=ending)
    {
        if(c<0) return -1;
        if(c==clear)
        {
            curr_size=size+1;
            slot=newcodes;
            top_slot=1<<curr_size;
            while((c=get_next_code(f))==clear) ;
            if(c==ending) break;
            if(c>=slot) c=0;
            oc=fc=c;
            *sp++=c;
        }
        else
        {
            code=c;
            if(code>=slot)
            {
                code=oc;
                *sp++=fc;
            }
            while(code>=newcodes)
            {
                *sp++=suffix[code];
                code=prefix[code];
            }
            *sp++=code;
            if(slot<top_slot)
            {
                suffix[slot]=fc=code;
                prefix[slot++]=oc;
                oc=c;
            }
            if(slot>=top_slot)
            {
                if(curr_size<12)
                {
                    top_slot<<=1;
                    ++curr_size;
                }
            }
        }
        while(sp>dstack)
        {
            if(x>=width)
            {
                switch(interlaced)
                {
                case 0: y++;   if(y>=height) y=0;                    break;
                case 1: y+=8;  if(y>=height) { y=4; interlaced=2; }  break;
                case 2: y+=8;  if(y>=height) { y=2; interlaced=3; }  break;
                case 3: y+=4;  if(y>=height) { y=1; interlaced=4; }  break;
                case 4: y+=2;  if(y>=height) { y=0; interlaced=0; }  break;
                }
                ptr=data+skew*(fullheight-1-top-y)+left;
                x=0;
            }
            *ptr++=*(--sp);
            x++;
        }
    }
    return 0;
}

DIB *LoadGIF(char filename[])
{
    FILE *f;
    struct
    {
        char gif[3];
        char ver[3];
        unsigned short width;
        unsigned short height;
        unsigned char bitsperpixel:3; /* 1 dazuaddieren */
        unsigned char sortedpalette:1;
        unsigned char depth:3;
        unsigned char globalpalette:1;
        unsigned char background_color_index:8;
        unsigned char aspect_ratio:8;
    }
    gifheader;
//#if sizeof(gifheader)!=13
//#error Bad gifheader structure
//#endif
    struct
    {
        unsigned short left;
        unsigned short top;
        unsigned short width;
        unsigned short height;
        unsigned char bitperpixel:3; /* 1 dazuaddieren */
        unsigned char reserved:2;
        unsigned char sortedpalette:1;
        unsigned char interlaced:1;
        unsigned char localcolorpalette:1;
    }
    bitheader;
//#if sizeof(bitheader)!=9
//#error Bad bitheader structure
//#endif
    DWORD skew;
    int ch,i;
    int colors;
    DIB *dib;
    char name[81];
    RGBQUAD globalpalette[256];

    AnsiToOem(filename,name);
    f=fopen(name,"rb");
    dib=NULL;
    if(f)
    {
        if(fread(&gifheader,sizeof(gifheader),1,f)==1&&memcmp(gifheader.gif,"GIF",3)==0)
        {
            dib=(DIB *)malloc(sizeof(DIB)+256*sizeof(RGBQUAD));
            if(dib)
            {
                memset(dib,0,sizeof(DIB));
                dib->bmi.bmiHeader.biSize=40;
                dib->bmi.bmiHeader.biWidth=gifheader.width;
                dib->bmi.bmiHeader.biHeight=gifheader.height;
                dib->bmi.bmiHeader.biPlanes=1;
                dib->bmi.bmiHeader.biBitCount=8;
                dib->bmi.bmiHeader.biCompression=BI_RGB;
                skew=((dib->bmi.bmiHeader.biWidth*dib->bmi.bmiHeader.biBitCount+31UL)/32UL)*4UL;
                dib->bmi.bmiHeader.biSizeImage=skew*dib->bmi.bmiHeader.biHeight;
                if((dib->data=GlobalLock(GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT,dib->bmi.bmiHeader.biSizeImage)))!=NULL)
                {
                    if(gifheader.background_color_index!=0)
                    {
                        /* _hmemset(dib->data,gifheader.background_color_index,dib->bmi.bmiHeader.biSizeImage); */
                        char huge *from;
                        char huge *to;
                        DWORD i;
                        
                        from=dib->data+1;
                        *from=gifheader.background_color_index;
                        to=dib->data;
                        *to=gifheader.background_color_index;
                        for(i=0;i<dib->bmi.bmiHeader.biSizeImage;i++) *to++=*from++;
                    }
                    colors=0;
                    if(gifheader.globalpalette) /* global color palette */
                    {
                        colors=2<<gifheader.bitsperpixel;
                        for(i=0;i<colors;i++)
                        {
                            globalpalette[i].rgbRed=getc(f);
                            globalpalette[i].rgbGreen=getc(f);
                            globalpalette[i].rgbBlue=getc(f);
                        }
                    }
                    while((ch=getc(f))=='!'||ch==','||ch=='\0')
                    {
                        if(ch=='!') /* skip comments */
                        {
                            getc(f);
                            while((i=getc(f))>0) while(i-->0) getc(f);
                        }
                        else if(ch==',')
                        {
                            fread(&bitheader,sizeof(bitheader),1,f);
                            if(bitheader.left+bitheader.width>gifheader.width||bitheader.top+bitheader.height>gifheader.height)
                            {
                                message("GIF Teilbild passt nicht in Gesamtbild");
                                break;
                            }
                            if(bitheader.localcolorpalette) /* local color palette */
                            {
                                colors=2<<bitheader.bitperpixel;
                                for(i=0;i<colors;i++)
                                {
                                    dib->bmi.bmiColors[i].rgbRed=getc(f);
                                    dib->bmi.bmiColors[i].rgbGreen=getc(f);
                                    dib->bmi.bmiColors[i].rgbBlue=getc(f);
                                }
                            }
                            else if(gifheader.globalpalette)
                            {
                                colors=2<<gifheader.bitsperpixel;
                                _fmemcpy(dib->bmi.bmiColors,globalpalette,colors*sizeof(RGBQUAD));
                            }
                            else
                            {
                                message("keine Farbpalette definiert");
                                break;
                            }
                            dib->bmi.bmiHeader.biClrUsed=colors;
                            decoder(f,bitheader.left,bitheader.top,bitheader.width,bitheader.height,dib->data,(long)skew,gifheader.width,gifheader.height,bitheader.interlaced);
                        }
                    }
                }
                else
                {
                    free(dib);
                    dib=NULL;
                }
            }
        }
        fclose(f);
    }
    return dib;
}
