/* LOADIMG */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "dib.h"

#pragma pack(1)

#define MAKEWORD(lo,hi) ((unsigned int)(lo)|((unsigned int)(hi)<<8))

DIB *LoadIMG(char filename[])
{
    struct
    {
        unsigned char version_hi,version_lo;
        unsigned char headlen_hi,headlen_lo;
        unsigned char planes_hi,planes_lo;
        unsigned char patlen_hi,patlen_lo;
        unsigned char pixelwidth_hi,pixelwidth_lo;
        unsigned char pixelheight_hi,pixelheight_lo;
        unsigned char width_hi,width_lo;
        unsigned char height_hi,height_lo;
        unsigned char ximg[4]; /* 'XIMG' */
        unsigned char model_hi,model_lo;
    }
    header;
    static RGBQUAD defaultPalette[16]=
    {
        {0xFF,0xFF,0xFF,0}, /* weiss */
        {0x00,0x00,0xFF,0}, /* hellrot */
        {0x00,0xFF,0x00,0}, /* hellgrn */
        {0x00,0xFF,0xFF,0}, /* hellgelb */
        {0xFF,0x00,0x00,0}, /* hellblau */
        {0xFF,0x00,0xFF,0}, /* hellmagenta */
        {0xFF,0xFF,0x00,0}, /* hellcyan */
        {0xC0,0xC0,0xC0,0}, /* hellgrau */
        {0x80,0x80,0x80,0}, /* dunkelgrau */
        {0x00,0x00,0x80,0}, /* dunkelrot */
        {0x00,0x80,0x00,0}, /* dunkelgrn */
        {0x00,0x80,0x80,0}, /* dunkelgelb */
        {0x80,0x00,0x00,0}, /* dunkelblau */
        {0x80,0x00,0x80,0}, /* dunkelmagenta */
        {0x80,0x80,0x00,0}, /* dunkelcyan */
        {0x00,0x00,0x00,0} /* schwarz */
    };
    BYTE value;
    unsigned char hi,lo;
    char huge *ptr;
    DIB *dib;
    FILE *f;
    int scanlen;
    int i,n,j;
    short colors;
    DWORD skew;
    static BYTE bits[4][4]=
    {
        {0x00,0x01,0x10,0x11},
        {0x00,0x02,0x20,0x22},
        {0x00,0x04,0x40,0x44},
        {0x00,0x08,0x80,0x88},
    };
    int bitsperpixel;
    int planes;
    int pos;
    int length;
    int count;
    int repeat;
    static BYTE pattern[256];
    char name[81];

    AnsiToOem(filename,name);
    f=fopen(name,"rb");
    dib=NULL;
    if(f)
    {
        if(fread(&header,sizeof(header),1,f)==1)
        {
            if(MAKEWORD(header.planes_lo,header.planes_hi)==1)
            {
                bitsperpixel=1;
            }
            else if(MAKEWORD(header.planes_lo,header.planes_hi)<=4)
            {
                bitsperpixel=4;
            }
            else if(MAKEWORD(header.planes_lo,header.planes_hi)<=8)
            {
                bitsperpixel=8;
            }
            else /* header.planes<=24 hoffentlich */
            {
                bitsperpixel=24;
            }
            colors=1<<bitsperpixel; /* ergibt 0 bei 24 Bit */
            dib=(DIB *)malloc(sizeof(DIB)+(1<<bitsperpixel)*sizeof(RGBQUAD));
            if(dib)
            {
                memset(dib,0,sizeof(DIB));
                memcpy(dib->bmi.bmiColors,defaultPalette,min(colors,16)*sizeof(RGBQUAD));
                memset(&dib->bmi.bmiColors[colors-1],0,sizeof(RGBQUAD));
                dib->bmi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
                dib->bmi.bmiHeader.biWidth=MAKEWORD(header.width_lo,header.width_hi);
                dib->bmi.bmiHeader.biHeight=MAKEWORD(header.height_lo,header.height_hi);
                dib->bmi.bmiHeader.biPlanes=1;
                dib->bmi.bmiHeader.biBitCount=bitsperpixel;
                dib->bmi.bmiHeader.biCompression=BI_RGB;
                planes=MAKEWORD(header.planes_lo,header.planes_hi);
                dib->bmi.bmiHeader.biClrUsed=1<<planes;
                scanlen=(MAKEWORD(header.width_lo,header.width_hi)+7)/8;
                skew=((dib->bmi.bmiHeader.biWidth*dib->bmi.bmiHeader.biBitCount+31)/32)*4;
                dib->bmi.bmiHeader.biSizeImage=(long)skew*dib->bmi.bmiHeader.biHeight;
                if(memcmp(header.ximg,"XIMG",4)==0)
                {
                    for(i=0;i<colors;i++)
                    {
                        hi=getc(f); lo=getc(f);
                        dib->bmi.bmiColors[i].rgbRed=MulDiv(MAKEWORD(lo,hi),255,1000);
                        hi=getc(f); lo=getc(f);
                        dib->bmi.bmiColors[i].rgbGreen=MulDiv(MAKEWORD(lo,hi),255,1000);
                        hi=getc(f); lo=getc(f);
                        dib->bmi.bmiColors[i].rgbBlue=MulDiv(MAKEWORD(lo,hi),255,1000);
                        dib->bmi.bmiColors[i].rgbReserved=0;
                    }
                }
                fseek(f,2L*MAKEWORD(header.headlen_lo,header.headlen_hi),SEEK_SET);
                pos=count=length=repeat=0;
                if((dib->data=GlobalLock(GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT,dib->bmi.bmiHeader.biSizeImage)))!=NULL)
                {
                    count=0;
                    for(j=(int)dib->bmi.bmiHeader.biHeight;j-->0;)
                    {
                        if(repeat>0)
                        {
                            ptr=dib->data+(long)skew*j;
                            for(i=0;i<(int)skew;i++,ptr++) ptr[0]=ptr[skew];
                            repeat--;
                        }
                        else for(n=0;n<planes;n++)
                        {
                            ptr=dib->data+(long)skew*j+n/8;
                            for(i=0;i<scanlen;i++)
                            {
                                if(pos>=length)
                                {
                                    pos=0;
                                    count--;
                                    while(count<=0)
                                    {
                                        value=getc(f);
                                        if(value==0x00)
                                        {
                                            value=getc(f);
                                            if(value==0x00)
                                            {
                                                value=getc(f); /* 0xFF berspringen */
                                                repeat=getc(f);
                                            }
                                            else
                                            {
                                                fread(pattern,length=MAKEWORD(header.patlen_lo,header.patlen_hi),1,f);
                                                count=value;
                                            }
                                        }
                                        else if(value==0x80)
                                        {
                                            length=getc(f);
                                            count=1;
                                            fread(pattern,1,length,f);
                                        }
                                        else
                                        {
                                            length=1;
                                            count=value&0x7F;
                                            pattern[0]=value&0x80?0xFF:0x00;
                                        }
                                    }
                                }
                                value=pattern[pos++];
                                if(dib->bmi.bmiHeader.biBitCount!=4)
                                {
                                    *ptr++=value;
                                }
                                else
                                {
                                    *ptr++|=bits[n][(value>>6)&3];
                                    *ptr++|=bits[n][(value>>4)&3];
                                    *ptr++|=bits[n][(value>>2)&3];
                                    *ptr++|=bits[n][value&3];
                                }
                            }
                        }
                    }
                }
                else
                {
                    free(dib);
                    dib=NULL;
                }
            }
        }
        fclose(f);
    }
    return dib;
}
