/* LOADTGA */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include "dib.h"

#pragma pack(1)

#define MAKEWORD(lo,hi) ((unsigned int)(lo)|((unsigned int)(hi)<<8))

DIB *LoadTGA(char filename[])
{
    struct
    {
        unsigned char IDLength;   /* Length of Identifier String */
        unsigned char CoMapType;  /* 0 = NoMap, other = RGB */
        unsigned char ImgType;    /* Image Type (1,2,3,9,10,11) */
        unsigned char Index_lo, Index_hi;   /* Index of first colour map entry */
        unsigned char Length_lo, Length_hi; /* Length of colour map (number of entries) */
        unsigned char CoSize;     /* Length of colour map entry (8,15,16,24,32) */
        unsigned char X_org_lo, X_org_hi;   /* X Origin of Image */
        unsigned char Y_org_lo, Y_org_hi;   /* Y Origin of Image */
        unsigned char Width_lo, Width_hi;   /* Width of Image */
        unsigned char Height_lo, Height_hi; /* Height of Image */
        unsigned char PixelSize;  /* Pixel Size (8,15,16,24,32) */
        unsigned char AttBits: 4; /* Number of Attribute Bits per pixel */
        unsigned char Rsrvd: 1;   /* Reserved bit */
        unsigned char OrgBit: 1;  /* Origin Bit (0=lower left, 1=upper left) */
        unsigned char IntrLve: 2; /* Interleaving Flag */
    }
    header;
    typedef struct { unsigned int b:5,g:5,r:5; } TGA16;
    unsigned int width,height,x,y,i,colors;
    TGA16 pixel16;
    char huge *ptr;
    FILE *f;
    long skew,line;
    unsigned char r,g,b,runlen;
    DIB *dib;
    char name[81];

    AnsiToOem(filename,name);
    f=fopen(name,"rb");
    dib=NULL;
    if(f)
    {
        if(fread(&header,sizeof(header),1,f)==1
        && (  (header.ImgType==1||header.ImgType==9)&&header.PixelSize==8&&header.CoMapType!=0&&header.CoSize!=0
           || (header.ImgType==2||header.ImgType==10)&&header.PixelSize!=8
           || (header.ImgType==3||header.ImgType==11)&&header.PixelSize==8
           )
        && header.IntrLve!=3
        &&(header.PixelSize==8||header.PixelSize==15||header.PixelSize==16||
           header.PixelSize==24||header.PixelSize==32)
        &&(header.CoSize==0||header.CoSize==8||header.CoSize==15||
           header.CoSize==16||header.CoSize==24||header.CoSize==32))
        {
            fseek(f,header.IDLength,SEEK_CUR);
            switch(header.ImgType)
            {
            case 1:
            case 9:
                colors=MAKEWORD(header.Length_lo,header.Length_hi);
                break;
            case 2:
            case 10:
                colors=0;
                break;
            case 3:
            case 11:
                colors=256;
                break;
            }
            dib=(DIB*)malloc(sizeof(DIB)+colors*sizeof(RGBQUAD));
            if(dib)
            {
                memset(dib,0,sizeof(DIB)+colors*sizeof(RGBQUAD));
                dib->bmi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
                dib->bmi.bmiHeader.biWidth=width=MAKEWORD(header.Width_lo,header.Width_hi);
                dib->bmi.bmiHeader.biHeight=height=MAKEWORD(header.Height_lo,header.Height_hi);
                dib->bmi.bmiHeader.biPlanes=1;
                dib->bmi.bmiHeader.biBitCount=header.PixelSize==8?8:24;
                dib->bmi.bmiHeader.biCompression=BI_RGB;
                dib->bmi.bmiHeader.biClrUsed=colors;
                switch(header.ImgType)
                {
                case 1:
                case 9:
                    for(i=0;i<colors;i++)
                    {
                        switch(header.CoSize)
                        {
                        case 8:
                            r=g=b=getc(f);
                            break;
                        case 15:
                        case 16:
                            *(int *)&pixel16=getw(f);
                            r=pixel16.r<<3;
                            g=pixel16.g<<3;
                            b=pixel16.b<<3;
                            break;
                        case 24:
                            b=getc(f);
                            g=getc(f);
                            r=getc(f);
                            break;
                        case 32:
                            b=getc(f);
                            g=getc(f);
                            r=getc(f);
                            getc(f);
                            break;
                        }
                        dib->bmi.bmiColors[i].rgbRed=r;
                        dib->bmi.bmiColors[i].rgbGreen=g;
                        dib->bmi.bmiColors[i].rgbBlue=b;
                        dib->bmi.bmiColors[i].rgbReserved=0;
                    }
                    break;
                case 3:
                case 11:
                    for(i=0;i<256;i++)
                    {
                        dib->bmi.bmiColors[i].rgbRed=i;
                        dib->bmi.bmiColors[i].rgbGreen=i;
                        dib->bmi.bmiColors[i].rgbBlue=i;
                        dib->bmi.bmiColors[i].rgbReserved=0;
                    }
                    break;
                }
                skew=((dib->bmi.bmiHeader.biWidth*dib->bmi.bmiHeader.biBitCount+31)/32)*4;
                dib->bmi.bmiHeader.biSizeImage=skew*dib->bmi.bmiHeader.biHeight;
                if((dib->data=GlobalLock(GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT,dib->bmi.bmiHeader.biSizeImage)))!=NULL)
                {
                    runlen=header.ImgType<8;
                    for(y=0;y<height;y++)
                    {
                        line=(1L<<header.IntrLve)*y;
                        while(line>=(long)height) line-=height-1;
                        if(header.OrgBit) line=height-1-line;
                        ptr=dib->data+skew*line;
                        for(x=0;x<width;x++)
                        {
                            if(header.ImgType<8||runlen--<129)
                            {
                                if((runlen&0x7F)==0) runlen=getc(f)+1;
                                switch(header.PixelSize)
                                {
                                case 8:
                                    *ptr++=getc(f);
                                    break;
                                case 15:
                                case 16:
                                    *(int *)&pixel16=getw(f);
                                    r=pixel16.r<<3;
                                    g=pixel16.g<<3;
                                    b=pixel16.b<<3;
                                    break;
                                case 24:
                                    b=getc(f);
                                    g=getc(f);
                                    r=getc(f);
                                    break;
                                case 32:
                                    b=getc(f);
                                    g=getc(f);
                                    r=getc(f);
                                    getc(f);
                                    break;
                                }
                                if(header.PixelSize!=8)
                                {
                                    *ptr++=b;
                                    *ptr++=g;
                                    *ptr++=r;
                                }
                            }
                        }
                    }
                }
                else
                {
                    free(dib);
                    dib=NULL;
                }
            }
        }
        fclose(f);
    }
    return dib;
}
