/* PUZZLE.C */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <windows.h>
#include <commdlg.h>
#include <malloc.h>
#include "dib.h"

HANDLE shell;
void (FAR PASCAL *DragAcceptFiles)(HWND wnd,BOOL fAccept);
UINT (FAR PASCAL *DragQueryFile)(HANDLE hDrop,int iFile,LPSTR lpszFile,UINT cb);

extern DIB *LoadIFF(char filename[]);
extern DIB *LoadGIF(char filename[]);
extern DIB *LoadPCX(char filename[]);
extern DIB *LoadIMG(char filename[]);
extern DIB *LoadTGA(char filename[]);
extern DIB *LoadBMP(char filename[]);
extern DIB *LoadFLT(char filename[],HWND wnd,HANDLE hInst,char filtername[]);
extern DIB *RGBto256(DIB *dib);

int width=64; /* Gre der Puzzleteile (ohne berhang) */
int height=64;
int fullwidth=96;
int fullheight=96;
BOOL addresid=0;

typedef struct part
{
    int x,y; /* Position auf Bildschirm */
    int px,py; /* Position in picture */
    HBITMAP image;
    HBITMAP mask;
    unsigned rot:2;
    unsigned seq:1;
    unsigned lm:3,tm:3,rm:3,bm:3; /* Art der Rnder */
}
PART;

PART *part;
HBITMAP back;
int parts;
HPALETTE palette;
char helpfile[81];
char progname[]="Puzzle";
HBITMAP picture;
HBITMAP masks[2][4][5];
int planes,bitspixel;
int xoffset,yoffset;
RECT border;
BOOL hello=FALSE;
char picname[81];
char fltname[81];
int pictype;
HANDLE hFrame;
HANDLE hInst;
int sizeid=0;
HANDLE hAccel;
DWORD color[]={RGB(255,255,255),RGB(0,0,0),RGB(127,0,0),RGB(0,127,0),RGB(0,0,127),RGB(255,255,0),RGB(127,0,127),RGB(0,255,255),RGB(128,128,128),RGB(192,192,192)};
int backcolor=0;
BOOL smallones;

#pragma pack(1)

BOOL FAR PASCAL About(HWND hwnd,UINT msg,UINT wParam,long lParam)
{
    switch(msg)
    {
    case WM_INITDIALOG:
        return TRUE;
    case WM_COMMAND:
        EndDialog(hwnd,0);
        return TRUE;
    }
    return FALSE;
}

void WaitCursor(BOOL on)
{
    static int wait=0;
    POINT p;
    HWND w;
    HCURSOR h;

    if(on)
    {
        if(wait++==0)
        {
            SetCursor(LoadCursor(0,MAKEINTRESOURCE(IDC_WAIT)));
        }
    }
    else
    {
        if(wait--==1)
        {
            GetCursorPos(&p);
            w=WindowFromPoint(p);
            if(w)
            {
#ifdef WIN32
                h=(HCURSOR)GetClassLong(w,GCL_HCURSOR);
#else
                h=(HCURSOR)GetClassWord(w,GCW_HCURSOR);
#endif
            }
            else
            {
                h=LoadCursor(0,MAKEINTRESOURCE(IDC_ARROW));
            }
            SetCursor(h);
        }
    }
}

// Funktion zur Ausgabe von Programmablaufmeldungen in das Debug-Fenster
void Debug(char *format,...)
{
    char buffer[1024];
    va_list arg;

    va_start(arg,format);
    vsprintf(buffer,format,arg);
    va_end(arg);
    OutputDebugString(buffer);
}

void message(LPCSTR format,...)
{
    va_list arg;
    char line[256];
    char buffer[256];

    va_start(arg,format);
    if(!HIWORD(format))
    {
        LoadString(hInst,LOWORD(format),buffer,sizeof(buffer));
        format=buffer;
    }
    wvsprintf(line,format,arg);
    va_end(arg);
    MessageBox(hFrame,line,progname,MB_OK);
}

BOOL FAR PASCAL SetColor(HWND hwnd,UINT msg,UINT wParam,long lParam)
{
    static int i;
    HBRUSH hbr;
    HPEN hpen;

    switch(msg)
    {
    case WM_INITDIALOG:
        i=100+backcolor;
        return TRUE;
    case WM_DRAWITEM:
#define dis ((LPDRAWITEMSTRUCT)lParam)
        hpen=SelectObject(dis->hDC,CreatePen(PS_INSIDEFRAME,2,(int)dis->CtlID==i||dis->itemState&ODS_SELECTED?RGB(0,0,0):RGB(255,255,255)));
        hbr=SelectObject(dis->hDC,CreateSolidBrush(color[dis->CtlID-100]));
        Rectangle(dis->hDC,dis->rcItem.left,dis->rcItem.top,dis->rcItem.right,dis->rcItem.bottom);
        DeleteObject(SelectObject(dis->hDC,hbr));
        DeleteObject(SelectObject(dis->hDC,hpen));
        if(dis->itemState&ODS_FOCUS)
        {
            DrawFocusRect(dis->hDC,&dis->rcItem);
        }
        return 0L;
#undef dis
    case WM_COMMAND:
        switch(LOWORD(wParam))
        {
        case 1:
            backcolor=i-100;
            EndDialog(hwnd,1);
            return TRUE;
        case 2:
            EndDialog(hwnd,0);
            return TRUE;
        case 100:
        case 101:
        case 102:
        case 103:
        case 104:
        case 105:
        case 106:
        case 107:
        case 108:
        case 109:
#ifdef WIN32
            if(HIWORD(wParam)==BN_CLICKED)
#else
            if(HIWORD(lParam)==BN_CLICKED)
#endif
            {
                InvalidateRect(GetDlgItem(hwnd,i),NULL,FALSE);
                i=LOWORD(wParam);
                InvalidateRect(GetDlgItem(hwnd,i),NULL,FALSE);
                return TRUE;
            }
            break;
        }
    }
    return FALSE;
}

int Dialog(HWND hwnd,LPCSTR name,BOOL (FAR PASCAL *func)(HWND,UINT,UINT,long))
{
    FARPROC lpProc;
    int result;

    lpProc=MakeProcInstance(func,hInst);
    result=DialogBox(hInst,name,hwnd,lpProc);
    FreeProcInstance(lpProc);
    return result;
}

void FreeDIB(DIB *dib)
{
    if(dib)
    {
        if(dib->data)
        {
            GlobalUnlock(SEGMENT(dib->data));
            GlobalFree(SEGMENT(dib->data));
        }
        free(dib);
    }
}

void FreePicture(void)
{
    int i;

    if(picture) DeleteObject(picture);
    picture=0;
    if(palette) DeleteObject(palette);
    palette=0;
    if(part)
    {
        for(i=0;i<parts;i++)
        {
            if(part[i].mask) DeleteObject(part[i].mask);
            if(part[i].image) DeleteObject(part[i].image);
        }
        free(part);
        part=NULL;
        parts=0;
    }
}

HBITMAP MakeMask(HDC dc1,HDC dc2,PART *p)
{
    HBITMAP oldbm1,oldbm2,bm;

    bm=CreateBitmap(fullwidth,fullheight,1,1,NULL);
    if(bm)
    {
        oldbm1=SelectObject(dc1,bm);
        oldbm2=SelectObject(dc2,masks[sizeid][(p->rot)%4][p->tm]);
        BitBlt(dc1,0,0,fullwidth,fullheight,dc2,0,0,SRCCOPY);
        SelectObject(dc2,masks[sizeid][(p->rot+1)%4][p->rm]);
        BitBlt(dc1,0,0,fullwidth,fullheight,dc2,0,0,SRCAND);
        SelectObject(dc2,masks[sizeid][(p->rot+2)%4][p->bm]);
        BitBlt(dc1,0,0,fullwidth,fullheight,dc2,0,0,SRCAND);
        SelectObject(dc2,masks[sizeid][(p->rot+3)%4][p->lm]);
        BitBlt(dc1,0,0,fullwidth,fullheight,dc2,0,0,SRCAND);
        SelectObject(dc2,oldbm2);
        SelectObject(dc1,oldbm1);
    }
    return bm;
}

#define PIXEL(x,y) ptr[(y)*skew+(x)]

void RotatePart(HDC dc1,HDC dc2,PART *p,int i)
{
    int x,y;
    BITMAPINFO *bmi;
    BITMAP b;
    HBITMAP oldbm1,oldbm2;
    char far *data;
    unsigned int skew;

    p->rot=(p->rot+i)%4;
    if(i||!p->mask)
    {
        if(p->mask) DeleteObject(p->mask);
        p->mask=MakeMask(dc1,dc2,p);
    }
    if(i)
    {
        GetObject(p->image,sizeof(b),&b);
        bmi=malloc(sizeof(BITMAPINFOHEADER)+((1<<b.bmPlanes*b.bmBitsPixel)&255)*sizeof(RGBQUAD));
        if(bmi)
        {
            memset(bmi,0,sizeof(BITMAPINFOHEADER));
            bmi->bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
            bmi->bmiHeader.biWidth=fullwidth;
            bmi->bmiHeader.biHeight=fullheight;
            bmi->bmiHeader.biPlanes=1;
            bmi->bmiHeader.biBitCount=b.bmPlanes*b.bmBitsPixel<8?8:b.bmPlanes*b.bmBitsPixel;
            skew=((fullwidth*bmi->bmiHeader.biBitCount+31)/32)*4;
            bmi->bmiHeader.biSizeImage=fullheight*skew;
            bmi->bmiHeader.biClrUsed=bmi->bmiHeader.biClrImportant=(1<<b.bmPlanes*b.bmBitsPixel)&255;
            data=_fmalloc((size_t)bmi->bmiHeader.biSizeImage);
            if(data)
            {
                GetDIBits(dc1,p->image,0,fullheight,data,bmi,DIB_RGB_COLORS);
                if(bmi->bmiHeader.biBitCount<=8)
                {
                    char pixel;
                    char far *ptr;

                    ptr=data;
                    switch(i)
                    {
                    case 1:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=pixel;
                        }
                        break;
                    case 2:
                        for(y=0;y<fullheight;y++) for(x=0;x<fullwidth/2;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=pixel;
                        }
                        break;
                    case 3:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=pixel;
                        }
                        break;
                    }
                }
                else if(bmi->bmiHeader.biBitCount<=16)
                {
                    WORD pixel;
                    WORD far *ptr;

                    ptr=(WORD far *)data;
                    skew/=2;
                    switch(i)
                    {
                    case 1:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=pixel;
                        }
                        break;
                    case 2:
                        for(y=0;y<fullheight;y++) for(x=0;x<fullwidth/2;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=pixel;
                        }
                        break;
                    case 3:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=pixel;
                        }
                        break;
                    }
                }
                else if(bmi->bmiHeader.biBitCount<=24)
                {
                    typedef struct
                    {
                        unsigned char b,g,r;
                    }
                    RGB;
#if sizeof(RGB)!=3
#error Compile bytealigned!
#endif
                    RGB pixel;
                    RGB far *ptr;

                    ptr=(RGB far *)data;
                    skew/=3;
                    switch(i)
                    {
                    case 1:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=pixel;
                        }
                        break;
                    case 2:
                        for(y=0;y<fullheight;y++) for(x=0;x<fullwidth/2;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=pixel;
                        }
                        break;
                    case 3:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=pixel;
                        }
                        break;
                    }
                }
                else if(bmi->bmiHeader.biBitCount<=32)
                {
                    RGBQUAD pixel;
                    RGBQUAD far *ptr;

                    ptr=(RGBQUAD far *)data;
                    skew/=4;
                    switch(i)
                    {
                    case 1:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=pixel;
                        }
                        break;
                    case 2:
                        for(y=0;y<fullheight;y++) for(x=0;x<fullwidth/2;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=pixel;
                        }
                        break;
                    case 3:
                        for(y=0;y<fullheight/2;y++) for(x=y;x<fullwidth-1-y;x++)
                        {
                            pixel=PIXEL(x,y);
                            PIXEL(x,y)=PIXEL(y,fullheight-1-x);
                            PIXEL(y,fullheight-1-x)=PIXEL(fullwidth-1-x,fullheight-1-y);
                            PIXEL(fullwidth-1-x,fullheight-1-y)=PIXEL(fullwidth-1-y,x);
                            PIXEL(fullwidth-1-y,x)=pixel;
                        }
                        break;
                    }
                }
                SetDIBits(dc1,p->image,0,fullheight,data,bmi,DIB_RGB_COLORS);
                _ffree(data);
            }
            else Debug("no data");
            free(bmi);
        }
        else Debug("No bmi");
    }
    oldbm1=SelectObject(dc1,p->image);
    oldbm2=SelectObject(dc2,p->mask);
    BitBlt(dc1,0,0,fullwidth,fullheight,dc2,0,0,0x220326L);
    SelectObject(dc2,oldbm2);
    SelectObject(dc1,oldbm1);
}

void MakeBitmaps(HWND hwnd)
{
    HDC dc1,dc2;
    int i,n;
    HBITMAP oldbm1,oldbm2;
    HPALETTE oldpal1,oldpal2;

    WaitCursor(TRUE);
    dc1=CreateCompatibleDC(0);
    if(dc1)
    {
        if(palette) oldpal1=SelectPalette(dc1,palette,FALSE);
        oldbm1=SelectObject(dc1,picture);
        dc2=CreateCompatibleDC(0);
        if(dc2)
        {
            if(palette) oldpal2=SelectPalette(dc2,palette,FALSE);
            for(i=0;i<parts;i++)
            {
                part[i].image=CreateBitmap(fullwidth,fullheight,planes,bitspixel,NULL);
                if(!part[i].image)
                {
                    message(MAKEINTRESOURCE(1+addresid));
                    break;
                }
                oldbm2=SelectObject(dc2,part[i].image);
                BitBlt(dc2,0,0,fullwidth,fullheight,dc1,part[i].px-(fullwidth-width)/2,part[i].py-(fullheight-height)/2,SRCCOPY);
                SelectObject(dc2,oldbm2);
                n=part[i].rot;
                part[i].mask=0;
                part[i].rot=0;
                RotatePart(dc1,dc2,&part[i],n);
            }
            if(palette&&oldpal2) SelectPalette(dc2,oldpal2,FALSE);
            DeleteDC(dc2);
        }
        SelectObject(dc1,oldbm1);
        if(palette&&oldpal1) SelectPalette(dc1,oldpal1,FALSE);
        DeleteDC(dc1);
    }
    WaitCursor(FALSE);
}

HBITMAP BitmapFromDib(DIB *dib)
{
    LOGPALETTE *lp;
    HPALETTE oldpal;
    HBITMAP bm;
    HDC dc;
    int i;

    if(palette) DeleteObject(palette);
    palette=0;
    dc=GetDC(0);
    if(dc)
    {
        if(dib->bmi.bmiHeader.biPlanes*dib->bmi.bmiHeader.biBitCount==24&&GetDeviceCaps(dc,PLANES)*GetDeviceCaps(dc,BITSPIXEL)==8)
        {
            dib=RGBto256(dib);
        }
        if((short)dib->bmi.bmiHeader.biClrUsed)
        {
            lp=malloc(sizeof(LOGPALETTE)+(size_t)dib->bmi.bmiHeader.biClrUsed*sizeof(PALETTEENTRY));
            if(lp)
            {
                lp->palVersion=0x0300;
                lp->palNumEntries=(WORD)dib->bmi.bmiHeader.biClrUsed;
                for(i=0;i<lp->palNumEntries;i++)
                {
                    lp->palPalEntry[i].peRed=dib->bmi.bmiColors[i].rgbRed;
                    lp->palPalEntry[i].peGreen=dib->bmi.bmiColors[i].rgbGreen;
                    lp->palPalEntry[i].peBlue=dib->bmi.bmiColors[i].rgbBlue;
                    lp->palPalEntry[i].peFlags=PC_NOCOLLAPSE;
                }
                palette=CreatePalette(lp);
                if(palette)
                {
                    oldpal=SelectPalette(dc,palette,FALSE);
                    RealizePalette(dc); // wozu das ntig ist ?
                }
                free(lp);
            }
            else
            {
                message(MAKEINTRESOURCE(2+addresid),dib->bmi.bmiHeader.biClrUsed);
            }
        }
        bm=CreateDIBitmap(dc,&dib->bmi.bmiHeader,CBM_INIT,dib->data,&dib->bmi,DIB_RGB_COLORS);
        if(palette&&oldpal) SelectPalette(dc,oldpal,FALSE);
        ReleaseDC(0,dc);
    }
    FreeDIB(dib);
    return bm;
}

BOOL LoadPuzzle(HWND hwnd,LPSTR name)
{
    OPENFILENAME of;
    int n,k;
    DIB *dib;
    FILE *f;
    PART *p;
    static char szFile[81];
    char title[100];
    static char szFilter[]="Puzzle\0*.puz\0";
    BOOL result;

    result=FALSE;
    if(name)
    {
        lstrcpy(szFile,name);
    }
    else
    {
        memset(&of,0,sizeof(OPENFILENAME));
        szFile[0]='\0';
        of.lStructSize=sizeof(OPENFILENAME);
        of.hwndOwner=hwnd;
        of.hInstance=hInst;
        of.lpstrFilter=szFilter;
        of.nFilterIndex=1L;
        of.lpstrFile=szFile;
        of.nMaxFile=sizeof(szFile);
        of.lpstrTitle="Load Puzzle";
        of.lpstrDefExt="PUZ";
        of.Flags=OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST|OFN_HIDEREADONLY;
        if(!GetOpenFileName(&of)) return FALSE;
    }
    f=fopen(szFile,"rb");
    if(!f) return FALSE;
    WaitCursor(TRUE);
    k=getw(f); /* sizeid */
    n=getw(f); /* parts */
    p=malloc((n+1)*sizeof(PART));
    if(p)
    {
        fread(p,n,sizeof(PART),f);
        pictype=getw(f);
        fread(picname,sizeof(picname),1,f);
        fread(fltname,sizeof(fltname),1,f);
        switch(pictype)
        {
        case 1:
            dib=LoadBMP(picname);
            break;
        case 2:
            dib=LoadPCX(picname);
            break;
        case 3:
            dib=LoadIFF(picname);
            break;
        case 4:
            dib=LoadTGA(picname);
            break;
        case 5:
            dib=LoadIMG(picname);
            break;
        case 6:
            dib=LoadGIF(picname);
            break;
        default:
            dib=LoadFLT(picname,hwnd,hInst,fltname);
        }
        if(dib)
        {
            FreePicture();
            parts=n;
            part=p;
            sizeid=smallones=k;
            if(smallones)
            {
                fullwidth=fullheight=64;
                width=height=48;
            }
            else
            {
                fullwidth=fullheight=96;
                width=height=64;
            }
            border.left=border.top=(fullwidth-width)/2-1;
            border.right=border.left+(int)(dib->bmi.bmiHeader.biWidth/width)*width+2;
            border.bottom=border.top+(int)(dib->bmi.bmiHeader.biHeight/height)*height+2;
            picture=BitmapFromDib(dib);
            if(picture)
            {
                MakeBitmaps(hwnd);
                strcpy(title,"Puzzle");
                if(picture)
                {
                    strcat(title," - ");
                    strcat(title,szFile);
                }
                SetWindowText(hwnd,title);
                SetScrollRange(hwnd,SB_VERT,-1,1,FALSE);
                SetScrollPos(hwnd,SB_VERT,yoffset=0,TRUE);
                SetScrollRange(hwnd,SB_HORZ,-1,1,FALSE);
                SetScrollPos(hwnd,SB_HORZ,xoffset=0,TRUE);
                InvalidateRect(hwnd,NULL,TRUE);
                result=TRUE;
            }
            else
            {
                message(MAKEINTRESOURCE(3+addresid));
            }
        }
        else
        {
            message(MAKEINTRESOURCE(4+addresid));
        }
    }
    else
    {
        message(MAKEINTRESOURCE(5+addresid),parts);
    }
    fclose(f);
    WaitCursor(FALSE);
    return result;
}

BOOL SavePuzzle(HWND hwnd)
{
    OPENFILENAME of;
    FILE *f;
    static char szFile[81];
    static char szFilter[]="Puzzle\0*.puz\0";

    memset(&of,0,sizeof(OPENFILENAME));
    szFile[0]='\0';
    of.lStructSize=sizeof(OPENFILENAME);
    of.hwndOwner=hwnd;
    of.hInstance=hInst;
    of.lpstrFilter=szFilter;
    of.nFilterIndex=1L;
    of.lpstrFile=szFile;
    of.nMaxFile=sizeof(szFile);
    of.lpstrTitle="Save Puzzle";
    of.Flags=OFN_OVERWRITEPROMPT|OFN_PATHMUSTEXIST|OFN_HIDEREADONLY;
    of.lpstrDefExt="PUZ";
    if(!GetSaveFileName(&of)) return FALSE;
    f=fopen(szFile,"wb");
    if(!f) return FALSE;
    WaitCursor(TRUE);
    putw(sizeid,f);
    putw(parts,f);
    fwrite(part,parts,sizeof(PART),f);
    putw(pictype,f);
    fwrite(picname,sizeof(picname),1,f);
    fwrite(fltname,sizeof(fltname),1,f);
    fclose(f);
    WaitCursor(FALSE);
    return TRUE;
}

void DrawPart(HDC hdc,HDC dc,int x,int y,PART *p)
{
    HBITMAP oldbm;

    oldbm=SelectObject(dc,p->mask);
    BitBlt(hdc,p->x+x,p->y+y,fullwidth,fullheight,dc,0,0,SRCAND);
    SelectObject(dc,p->image);
    BitBlt(hdc,p->x+x,p->y+y,fullwidth,fullheight,dc,0,0,SRCPAINT);
    SelectObject(dc,oldbm);
}

void Draw(HDC hdc,int dx,int dy,RECT *clip)
{
    int i;
    HDC dc;
    RECT r;
    HPALETTE oldpal;
    HBRUSH hbr;

    CopyRect(&r,clip);
    OffsetRect(&r,dx,dy);
    hbr=CreateSolidBrush(color[backcolor]);
    FillRect(hdc,&r,hbr);
    DeleteObject(hbr);
    CopyRect(&r,&border);
    OffsetRect(&r,dx,dy);
    DrawFocusRect(hdc,&r);
    dc=CreateCompatibleDC(0);
    if(dc)
    {
        if(palette) oldpal=SelectPalette(dc,palette,FALSE);
        for(i=0;i<parts;i++)
        {
            SetRect(&r,part[i].x,part[i].y,part[i].x+fullwidth,part[i].y+fullwidth);
            if(IntersectRect(&r,&r,clip)) DrawPart(hdc,dc,dx,dy,&(part[i]));
        }
        if(palette&&oldpal) SelectPalette(dc,oldpal,FALSE);
        DeleteDC(dc);
    }
}

void AlignRect(HWND hwnd,RECT *r)
{
    POINT p;

    p.x=r->left+xoffset;
    p.y=0;
    ClientToScreen(hwnd,&p);
    r->left-=(p.x&15);
}

void DrawRect(HWND hwnd,RECT *r)
{
    HDC hdc,dc;
    HBITMAP bm,oldbm;
    HPALETTE oldpal,oldpal2;

    hdc=GetDC(hwnd);
    if(hdc)
    {
        if(palette)
        {
            oldpal=SelectPalette(hdc,palette,FALSE);
            RealizePalette(hdc);
        }
        AlignRect(hwnd,r);
        bm=CreateBitmap(r->right-r->left,r->bottom-r->top,planes,bitspixel,NULL);
        if(bm)
        {
            dc=CreateCompatibleDC(0);
            if(dc)
            {
                if(palette) oldpal2=SelectPalette(dc,palette,FALSE);
                oldbm=SelectObject(dc,bm);
                Draw(dc,-r->left,-r->top,r);
                BitBlt(hdc,r->left+xoffset,r->top+yoffset,r->right-r->left,r->bottom-r->top,dc,0,0,SRCCOPY);
                SelectObject(dc,oldbm);
                if(palette&&oldpal2) SelectPalette(dc,oldpal2,FALSE);
                DeleteDC(dc);
            }
            DeleteObject(bm);
        }
        if(!bm||!dc)
        {
            Draw(hdc,xoffset,yoffset,r);
        }
        if(palette&&oldpal) SelectPalette(hdc,oldpal,FALSE);
        ReleaseDC(hwnd,hdc);
    }
}

int zufall(unsigned x)
{
    return (unsigned)rand()%x;
}

void Scramble(HWND hwnd)
{
    RECT r;
    int i;

    GetClientRect(hwnd,&r);
    for(i=0;i<parts;i++)
    {
        part[i].x=zufall(r.right-r.left-width)-(fullwidth-width)/2;
        part[i].y=zufall(r.bottom-r.top-height)-(fullheight-height)/2;
        part[i].rot=zufall(4);
        part[i].seq=FALSE;
    }
    MakeBitmaps(hwnd);
}

void Solve(HWND hwnd)
{
    HDC dc1,dc2;
    HPALETTE oldpal1,oldpal2;
    int i;

    WaitCursor(TRUE);
    dc1=CreateCompatibleDC(0);
    if(dc1)
    {
        if(palette) oldpal1=SelectPalette(dc1,palette,FALSE);
        dc2=CreateCompatibleDC(0);
        if(dc2)
        {
            if(palette) oldpal2=SelectPalette(dc2,palette,FALSE);
            for(i=0;i<parts;i++)
            {
                part[i].x=part[i].px;
                part[i].y=part[i].py;
                RotatePart(dc1,dc2,&part[i],(4-part[i].rot)%4);
            }
            if(palette&&oldpal2) SelectPalette(dc2,oldpal2,FALSE);
            DeleteDC(dc2);
        }
        if(palette&&oldpal1) SelectPalette(dc1,oldpal1,FALSE);
        DeleteDC(dc1);
    }
    WaitCursor(FALSE);
}

void AddBounds(RECT *r,int n,int m)
{
    RECT t;

    while(n<m)
    {
        SetRect(&t,part[n].x,part[n].y,part[n].x+fullwidth,part[n].y+fullheight);
        UnionRect(r,r,&t);
        n++;
    }
}

void GetBounds(RECT *r,int n,int m)
{
    SetRect(r,part[n].x,part[n].y,part[n].x+fullwidth,part[n].y+fullheight);
    AddBounds(r,++n,m);
}

void PartBitmap(HWND hwnd)
{
    int match[]={0,2,1,4,3};
    int i,j,n;
    BITMAP bm;

    GetObject(picture,sizeof(BITMAP),(LPSTR)&bm);
    if(smallones)
    {
        fullwidth=fullheight=64;
        width=height=48;
    }
    else
    {
        fullwidth=fullheight=96;
        width=height=64;
    }
    sizeid=smallones;
    parts=(bm.bmWidth/width)*(bm.bmHeight/height);
    part=malloc((parts+1)*sizeof(PART));
    border.left=border.top=(fullwidth-width)/2-1;
    border.right=border.left+(int)(bm.bmWidth/width)*width+2;
    border.bottom=border.top+(int)(bm.bmHeight/height)*height+2;
    n=0;
    for(j=0;j<bm.bmHeight/height;j++)
    {
        for(i=0;i<bm.bmWidth/width;i++)
        {
            part[n].px=i*width;
            part[n].py=j*height;
            part[n].tm=j>0?match[part[n-bm.bmWidth/width].bm]:0;
            part[n].lm=i>0?match[part[n-1].rm]:0;
            part[n].bm=j<bm.bmHeight/height-1?zufall(4)+1:0;
            part[n].rm=i<bm.bmWidth/width-1?zufall(4)+1:0;
            n++;
        }
    }
    Scramble(hwnd);
}

BOOL LoadDIB(HWND hwnd,LPSTR name,int type)
{
    OPENFILENAME of;
    DIB *dib;
    static char szFile[81];
    char title[100];
    static char szFilter[]=
        "Bitmap(.BMP)\0*.bmp;*.dib;*.rle\0"
        "PC Paintbrush(.PCX)\0*.pcx;*.scr\0"
        "Electronic Arts(.IFF)\0*.lbm;*.iff\0"
        "Truevision Targa(.TGA)\0*.tga\0"
        "GEM Paint(.IMG)\0*.img\0"
        "Compuserve(.GIF)\0*.gif\0";
    static char buffer[4096];
    int i,len,n;
    char *comma;
    static char *filtersection="MS Graphic Import Filters";

    if(name)
    {
        lstrcpy(szFile,name);
        of.nFilterIndex=type;
    }
    else
    {
        memcpy(buffer,szFilter,sizeof(szFilter));
        n=sizeof(szFilter)-1;
        len=n+GetProfileString(filtersection,NULL,"",buffer+n,sizeof(buffer)-n);
        while(buffer[n])
        {
            GetProfileString(filtersection,buffer+n,"",title,sizeof(title));
            n+=strlen(buffer+n)+1;
            comma=strchr(title,',');
            if(comma)
            {
                i=strlen(comma+1)+1;
                memmove(buffer+n+2+i,buffer+n,len-n);
                memcpy(buffer+n,"*.",2);
                memcpy(buffer+n+2,comma+1,i);
                len+=i+2;
                n+=i+2;
            }
        }
        buffer[n]='\0';
        memset(&of,0,sizeof(OPENFILENAME));
        szFile[0]='\0';
        of.lStructSize=sizeof(OPENFILENAME);
        of.hwndOwner=hwnd;
        of.hInstance=hInst;
        of.lpstrFilter=buffer;
        of.nFilterIndex=1L;
        of.lpstrFile=szFile;
        of.nMaxFile=sizeof(szFile);
        of.lpstrTitle="Load Picture";
        of.Flags=OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST|OFN_HIDEREADONLY;
        if(!GetOpenFileName(&of)) return FALSE;
    }
    WaitCursor(TRUE);
    switch(of.nFilterIndex)
    {
    case 1:
        dib=LoadBMP(szFile);
        break;
    case 2:
        dib=LoadPCX(szFile);
        break;
    case 3:
        dib=LoadIFF(szFile);
        break;
    case 4:
        dib=LoadTGA(szFile);
        break;
    case 5:
        dib=LoadIMG(szFile);
        break;
    case 6:
        dib=LoadGIF(szFile);
        break;
    default:
        for(n=0,i=1;i<(int)of.nFilterIndex;i++)
        {
            n+=strlen(buffer+n)+1;
            n+=strlen(buffer+n)+1;
        }
        GetProfileString(filtersection,buffer+n,"",fltname,sizeof(fltname));
        comma=strchr(fltname,',');
        if(comma)
        {
            *comma='\0';
            dib=LoadFLT(szFile,hwnd,hInst,fltname);
        }
        else
        {
            dib=NULL;
            message(MAKEINTRESOURCE(6+addresid));
        }
    }
    if(!dib)
    {
        WaitCursor(FALSE);
        message(MAKEINTRESOURCE(7+addresid));
        return FALSE;
    }
    if(dib->bmi.bmiHeader.biWidth<width||dib->bmi.bmiHeader.biHeight<height)
    {
        WaitCursor(FALSE);
        message(MAKEINTRESOURCE(8+addresid));
        return FALSE;
    }
    FreePicture();
    picture=BitmapFromDib(dib);
    strcpy(title,"Puzzle");
    if(picture)
    {
        strcat(title," - ");
        strcat(title,szFile);
    }
    SetWindowText(hwnd,title);
    PartBitmap(hwnd);
    SetScrollRange(hwnd,SB_VERT,-1,1,FALSE);
    SetScrollPos(hwnd,SB_VERT,yoffset=0,TRUE);
    SetScrollRange(hwnd,SB_HORZ,-1,1,FALSE);
    SetScrollPos(hwnd,SB_HORZ,xoffset=0,TRUE);
    InvalidateRect(hwnd,NULL,TRUE);
    pictype=(int)of.nFilterIndex;
    strcpy(picname,szFile);
    WaitCursor(FALSE);
    return dib&&picture;
}

void LoadPicture(HWND hwnd,LPSTR name)
{
    int i;

    i=lstrlen(name);
    while(i>0&&name[i]!='\\'&&name[i]!='/'&&name[i]!=':') i--;
    while(name[i]!='\0'&&name[i]!='.') i++;
    if(lstrcmpi(name+i,".PUZ")==0)
    {
        LoadPuzzle(hwnd,name);
        return;
    }
    else if(lstrcmpi(name+i,".BMP")==0||lstrcmpi(name+i,".DIB")==0||lstrcmpi(name+i,".RLE")==0)
    {
        i=1;
    }
    else if(lstrcmpi(name+i,".LBM")==0||lstrcmpi(name+i,".IFF")==0)
    {
        i=3;
    }
    else if(lstrcmpi(name+i,".GIF")==0)
    {
        i=6;
    }
    else if(lstrcmpi(name+i,".PCX")==0||lstrcmpi(name+i,".SCR")==0)
    {
        i=2;
    }
    else if(lstrcmpi(name+i,".IMG")==0)
    {
        i=5;
    }
    else if(lstrcmpi(name+i,".TGA")==0)
    {
        i=4;
    }
    else
    {
        message(MAKEINTRESOURCE(9+addresid),name);
        return;
    }
    LoadDIB(hwnd,name,i);
    return;
}

void Scroll(HWND hwnd,int n,short dir,short value)
{
    int pos,oldpos,minpos,maxpos;
    RECT r;

    oldpos=pos=GetScrollPos(hwnd,n);
    GetClientRect(hwnd,&r);
    switch(dir)
    {
    case SB_LINEUP:
        pos-=16;
        break;
    case SB_LINEDOWN:
        pos+=16;
        break;
    case SB_PAGEUP:
        if(n==SB_HORZ)
        {
            pos-=(r.right-r.left)/2;
        }
        else
        {
            pos-=(r.bottom-r.top)/2;
        }
        break;
    case SB_PAGEDOWN:
        if(n==SB_HORZ)
        {
            pos+=(r.right-r.left)/2;
        }
        else
        {
            pos+=(r.bottom-r.top)/2;
        }
        break;
    case SB_THUMBPOSITION:
        pos=value;
        break;
    }
    GetScrollRange(hwnd,n,&minpos,&maxpos);
    if(pos<minpos)
    {
        SetScrollRange(hwnd,n,pos,maxpos,FALSE);
    }
    else if(pos>maxpos)
    {
        SetScrollRange(hwnd,n,minpos,pos,FALSE);
    }
    SetScrollPos(hwnd,n,pos,TRUE);
    if(pos!=oldpos)
    {
        if(n==SB_HORZ)
        {
            ScrollWindow(hwnd,oldpos-pos,0,NULL,NULL);
            xoffset=-pos;
        }
        else
        {
            ScrollWindow(hwnd,0,oldpos-pos,NULL,NULL);
            yoffset=-pos;
        }
        UpdateWindow(hwnd);
    }
}

void RotateBytes(void *ptr,int lo,int hi,int dist)
{
    int n,i,j,k;
    char ch;

    i=k=lo;
    ch=((char *)ptr)[i];
    for(n=lo;n<hi;n++)
    {
        if(dist>0)
        {
            j=i<lo+dist?i+(hi-lo)-dist:i-dist;
        }
        else
        {
            j=i>=hi+dist?i-(hi-lo)-dist:i-dist;
        }
        if(j==k)
        {
            ((char *)ptr)[i]=ch;
            i=++k;
            ch=((char *)ptr)[i];
        }
        else
        {
            ((char *)ptr)[i]=((char *)ptr)[j];
            i=j;
        }
    }
}

int Gather(int i)
{
    int n;

    for(n=1;i>0&&part[i-1].seq;i--) n++;
    while(i+n-1<parts&&part[i+n-1].seq) n++;
    RotateBytes(part,i*sizeof(PART),parts*sizeof(PART),-n*sizeof(PART));
    return n;
}

void Drop(HWND hwnd,int moving)
{
    int i,j,n,dx,dy;
    RECT r;

    if(parts==moving) return;
    for(i=0;;)
    {
        for(j=parts-moving;j<parts;j++)
        {
            if(part[i].rot==part[j].rot)
            {
                if(part[i].rot==0 &&
                (  part[i].px==part[j].px+width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-(part[j].x+width))<6 && abs(dy=part[i].y-part[j].y)<6
                || part[i].px==part[j].px-width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-(part[j].x-width))<6 && abs(dy=part[i].y-part[j].y)<6
                || part[i].px==part[j].px && part[i].py==part[j].py+height &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y+height))<6
                || part[i].px==part[j].px && part[i].py==part[j].py-height &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y-height))<6
                ) ) break;
                if(part[i].rot==1 &&
                (  part[i].px==part[j].px+width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y+height))<6
                || part[i].px==part[j].px-width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y-height))<6
                || part[i].px==part[j].px && part[i].py==part[j].py+height &&
                   abs(dx=part[i].x-(part[j].x-width))<6 && abs(dy=part[i].y-part[j].y)<6
                || part[i].px==part[j].px && part[i].py==part[j].py-height &&
                   abs(dx=part[i].x-(part[j].x+width))<6 && abs(dy=part[i].y-part[j].y)<6
                ) ) break;
                if(part[i].rot==2 &&
                (  part[i].px==part[j].px+width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-(part[j].x-width))<6 && abs(dy=part[i].y-part[j].y)<6
                || part[i].px==part[j].px-width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-(part[j].x+width))<6 && abs(dy=part[i].y-part[j].y)<6
                || part[i].px==part[j].px && part[i].py==part[j].py+height &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y-height))<6
                || part[i].px==part[j].px && part[i].py==part[j].py-height &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y+height))<6
                ) ) break;
                if(part[i].rot==3 &&
                (  part[i].px==part[j].px+width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y-height))<6
                || part[i].px==part[j].px-width && part[i].py==part[j].py &&
                   abs(dx=part[i].x-part[j].x)<6 && abs(dy=part[i].y-(part[j].y+height))<6
                || part[i].px==part[j].px && part[i].py==part[j].py+height &&
                   abs(dx=part[i].x-(part[j].x+width))<6 && abs(dy=part[i].y-part[j].y)<6
                || part[i].px==part[j].px && part[i].py==part[j].py-height &&
                   abs(dx=part[i].x-(part[j].x-width))<6 && abs(dy=part[i].y-part[j].y)<6
                ) ) break;
            }
        }
        if(j<parts)
        {
            for(n=1;i>0&&part[i-1].seq;i--) n++;
            while(i+n<parts-moving&&part[i+n-1].seq) n++;
            RotateBytes(part,(i+n)*sizeof(PART),parts*sizeof(PART),moving*sizeof(PART));
            part[i+n-1].seq=TRUE;
            n=i+n;
            break;
        }
        else
        {
            i++;
            if(i>=parts-moving)
            {
                n=parts-moving;
                if(part[i].rot==0&&abs(dx=part[i].px-part[i].x)<6&&abs(dy=part[i].py-part[i].y)<6) break;
                return;
            }
        }
    }
    GetBounds(&r,n,n+moving);
    for(j=n;j<n+moving;j++)
    {
        part[j].x+=dx;
        part[j].y+=dy;
    }
    AddBounds(&r,n,n+moving);
    DrawRect(hwnd,&r);
    MessageBeep(0);
    if(hello)
    {
        for(i=0;i<parts;i++) 
        {
            if(part[i].rot!=0||part[i].x!=part[i].px||part[i].y!=part[i].py) break;
        }
        if(i==parts) WriteProfileString(progname,"HelloDone","1");
    }
}

int FindPart(HWND hwnd,int x,int y)
{
    HDC dc;
    int i;
    DWORD m;
    HBITMAP oldbm;

    i=0;
    dc=CreateCompatibleDC(0);
    if(dc)
    {
        for(i=parts-1;i>=0;i--) if(part[i].x<=x&&part[i].y<=y&&x<part[i].x+fullwidth&&y<part[i].y+fullheight)
        {
            oldbm=SelectObject(dc,part[i].mask);
            m=GetPixel(dc,x-part[i].x,y-part[i].y);
            SelectObject(dc,oldbm);
            if(!m) break;
        }
        DeleteDC(dc);
    }
    return i;
}

int GrabParts(HWND hwnd,int x,int y)
{
    int i;

    i=FindPart(hwnd,x,y);
    if(i<0) return 0;
    return Gather(i);
}

void Hint(HWND hwnd)
{
    HDC dc1,dc2;
    HPALETTE oldpal1,oldpal2;
    int n,i,z,x,y,k;
    RECT r;

    n=Gather(parts-1);
    if(n)
    {
        if(part[parts-1].rot)
        {
            dc1=CreateCompatibleDC(0);
            if(dc1)
            {
                if(palette) oldpal1=SelectPalette(dc1,palette,FALSE);
                dc2=CreateCompatibleDC(0);
                if(dc2)
                {
                    if(palette) oldpal2=SelectPalette(dc2,palette,FALSE);
                    GetBounds(&r,parts-n,parts);
                    x=(r.right+r.left)/2;
                    y=(r.bottom+r.top)/2;
                    for(i=parts-n;i<parts;i++)
                    {
                        k=(4-part[i].rot)%4;
                        RotatePart(dc1,dc2,&part[i],k);
                        while(k-->0)
                        {
                            z=x+(y-(part[i].y+fullwidth/2))-fullwidth/2;
                            part[i].y=y-(x-(part[i].x+fullheight/2))-fullheight/2;
                            part[i].x=z;
                        }
                    }
                    AddBounds(&r,parts-n,parts);
                    if(palette&&oldpal2) SelectPalette(dc2,oldpal2,FALSE);
                    DeleteDC(dc2);
                }
                if(palette&&oldpal1) SelectPalette(dc1,oldpal1,FALSE);
                DeleteDC(dc1);
            }
            DrawRect(hwnd,&r);
        }
        while(part[parts-1].x!=part[parts-1].px||part[parts-1].y!=part[parts-1].py)
        {
            GetBounds(&r,parts-n,parts);
            for(i=parts-n;i<parts;i++)
            {
                if(part[i].x>part[i].px)
                {
                    part[i].x-=min(part[i].x-part[i].px,8*n);
                }
                else if(part[i].x<part[i].px)
                {
                    part[i].x+=min(part[i].px-part[i].x,8*n);
                }
                if(part[i].y>part[i].py)
                {
                    part[i].y-=min(part[i].y-part[i].py,8*n);
                }
                else if(part[i].y<part[i].py)
                {
                    part[i].y+=min(part[i].py-part[i].y,8*n);
                }
            }
            AddBounds(&r,parts-n,parts);
            DrawRect(hwnd,&r);
        }
    }
}

long FAR PASCAL MainWndProc(HWND hwnd,UINT msg,UINT wParam,LONG lParam)
{
    PAINTSTRUCT ps;
    int i;
    static int lastx,lasty;
    HCURSOR hc;
    RECT r;
    HBRUSH hbr;
    HPALETTE oldpal;
    HDC dc;
    static int moving;
    int rotating;
    static int timer;

    switch(msg)
    {
    case WM_CREATE:
        dc=CreateCompatibleDC(0);
        planes=GetDeviceCaps(dc,PLANES);
        bitspixel=GetDeviceCaps(dc,BITSPIXEL);
        DeleteDC(dc);
        back=CreateBitmap(1,1,planes,bitspixel,NULL);
        SetScrollRange(hwnd,SB_VERT,0,0,FALSE);
        SetScrollRange(hwnd,SB_HORZ,0,0,FALSE);
        i=SetErrorMode(SEM_NOOPENFILEERRORBOX);
        shell=LoadLibrary("SHELL.DLL");
        SetErrorMode(i);
        SetFocus(hwnd);
        if(shell>=32)
        {
            (FARPROC)DragAcceptFiles=GetProcAddress(shell,"DragAcceptFiles");
            if(DragAcceptFiles)
            {
                DragAcceptFiles(hwnd,TRUE);
                (FARPROC)DragQueryFile=GetProcAddress(shell,"DragQueryFile");
            }
        }
        break;
    case WM_DROPFILES:
        if(shell>=32&&DragQueryFile)
        {
            char name[81];

            DragQueryFile((HANDLE)wParam,0,name,sizeof(name));
            LoadPicture(hwnd,name);
            if(IsIconic(hwnd))
            {
                ShowWindow(hwnd,SW_RESTORE);
            }
            else
            {
                BringWindowToTop(hwnd);
            }
        }
        return 0L;
    case WM_INITMENU:
        if((HMENU)wParam==GetMenu(hwnd))
        {
            CheckMenuItem((HMENU)wParam,101,smallones?MF_CHECKED:MF_UNCHECKED);
            EnableMenuItem((HMENU)wParam,103,picname[0]?MF_ENABLED:MF_GRAYED);
            EnableMenuItem((HMENU)wParam,106,parts?MF_ENABLED:MF_GRAYED);
            EnableMenuItem((HMENU)wParam,107,parts?MF_ENABLED:MF_GRAYED);
            EnableMenuItem((HMENU)wParam,108,parts&&(part[parts-1].x!=part[parts-1].px||part[parts-1].y!=part[parts-1].py)?MF_ENABLED:MF_GRAYED);
            EnableMenuItem((HMENU)wParam,109,parts?MF_ENABLED:MF_GRAYED);
            return 0L;
        }
        break;
    case WM_KEYDOWN:
        if(wParam==VK_F1) WinHelp(hwnd,helpfile,HELP_INDEX,1L);
        return 0L;
    case WM_COMMAND:
        switch(LOWORD(wParam))
        {
        case 100:
            if(LoadDIB(hwnd,NULL,0))
            {
                hello=FALSE;
            }
            return TRUE;
        case 102:
            if(LoadPuzzle(hwnd,NULL))
            {
                hello=FALSE;
            }
            return TRUE;
        case 103:
            SavePuzzle(hwnd);
            return TRUE;
        case 104:
            DestroyWindow(hwnd);
            return TRUE;
        case 101:
            smallones=!smallones;
            break;
        case 106:
            if(parts)
            {
                Scramble(hwnd);
                SetScrollRange(hwnd,SB_VERT,-1,1,FALSE);
                SetScrollPos(hwnd,SB_VERT,yoffset=0,TRUE);
                SetScrollRange(hwnd,SB_HORZ,-1,1,FALSE);
                SetScrollPos(hwnd,SB_HORZ,xoffset=0,TRUE);
                InvalidateRect(hwnd,NULL,TRUE);
            }
            return TRUE;
        case 107:
            if(parts)
            {
                Solve(hwnd);
                InvalidateRect(hwnd,NULL,TRUE);
            }
            return TRUE;
        case 108:
            if(parts) Hint(hwnd);
            break;
        case 105:
            if(Dialog(hwnd,MAKEINTRESOURCE(3+addresid),SetColor))
            {
                InvalidateRect(hwnd,NULL,TRUE);
            }
            break;
        case 109:
            if(parts)
            {
                for(i=0;i<parts/2;i++)
                {
                    PART temp;

                    memcpy(&temp,part+i,sizeof(PART));
                    memcpy(part+i,part+parts-1-i,sizeof(PART));
                    memcpy(part+parts-1-i,&temp,sizeof(PART));
                }
                for(i=0;i<parts-1;i++)
                {
                    if(part[i].seq&&!part[i+1].seq)
                    {
                        part[i].seq=FALSE;
                    }
                    else if(!part[i].seq&&part[i+1].seq)
                    {
                        part[i].seq=TRUE;
                    }
                }
                part[i].seq=FALSE;
                InvalidateRect(hwnd,NULL,TRUE);
            }
            return 0L;
        case 200:
            Dialog(hwnd,MAKEINTRESOURCE(1+addresid),About);
            break;
        case 201:
            WinHelp(hwnd,helpfile,HELP_INDEX,0L);
            break;
        case 202:
            WinHelp(hwnd,helpfile,HELP_HELPONHELP,0L);
            break;
        }
        break;
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
        rotating=0;
        if(moving||(rotating=GrabParts(hwnd,LOWORD(lParam)-xoffset,HIWORD(lParam)-yoffset))>0)
        {
            RECT r;
            int x,y,i,z,k;
            HDC dc1,dc2;
            HPALETTE oldpal1,oldpal2;

            x=LOWORD(lParam)-xoffset;
            y=HIWORD(lParam)-yoffset;
            GetBounds(&r,parts-moving-rotating,parts);
            dc1=CreateCompatibleDC(0);
            if(dc1)
            {
                if(palette) oldpal1=SelectPalette(dc1,palette,FALSE);
                dc2=CreateCompatibleDC(0);
                if(dc2)
                {
                    if(palette) oldpal2=SelectPalette(dc2,palette,FALSE);
                    hc=SetCursor(LoadCursor(hInst,"Rota"));
                    for(i=parts-moving-rotating;i<parts;i++)
                    {
                        k=msg==WM_RBUTTONDOWN?1:3;
                        RotatePart(dc1,dc2,&part[i],k);
                        while(k-->0)
                        {
                            z=x+(y-(part[i].y+fullwidth/2))-fullwidth/2;
                            part[i].y=y-(x-(part[i].x+fullheight/2))-fullheight/2;
                            part[i].x=z;
                        }
                    }
                    SetCursor(hc);
                    if(palette&&oldpal2) SelectPalette(dc2,oldpal2,FALSE);
                    DeleteDC(dc2);
                }
                if(palette&&oldpal1) SelectPalette(dc1,oldpal1,FALSE);
                DeleteDC(dc1);
            }
            AddBounds(&r,parts-moving-rotating,parts);
            DrawRect(hwnd,&r);
        }
        return 0L;
    case WM_LBUTTONDOWN:
        lastx=LOWORD(lParam)-xoffset;
        lasty=HIWORD(lParam)-yoffset;
        moving=GrabParts(hwnd,lastx,lasty);
        if(moving>0)
        {
            GetBounds(&r,parts-moving,parts);
            DrawRect(hwnd,&r);
            SetCursor(LoadCursor(hInst,"Cross"));
            timer=SetTimer(hwnd,99,500,NULL);
        }
        SetCapture(hwnd); /* Mauskontrolle bernehmen */
        return 0L;
    case WM_MOUSEMOVE:
        if(moving>0)
        {
            RECT r1,r2,r;
            int i;

            SetCursor(LoadCursor(hInst,"Cross"));
            GetBounds(&r1,parts-moving,parts);
            for(i=parts-moving;i<parts;i++)
            {
                part[i].x+=LOWORD(lParam)-xoffset-lastx;
                part[i].y+=HIWORD(lParam)-yoffset-lasty;
            }
            lastx=LOWORD(lParam)-xoffset;
            lasty=HIWORD(lParam)-yoffset;
            GetBounds(&r2,parts-moving,parts);
            if(IntersectRect(&r,&r1,&r2))
            {
                UnionRect(&r,&r1,&r2);
                DrawRect(hwnd,&r);
            }
            else
            {
                DrawRect(hwnd,&r1);
                DrawRect(hwnd,&r2);
            }
        }
        else if(FindPart(hwnd,LOWORD(lParam)-xoffset,HIWORD(lParam)-yoffset)>=0)
        {
            SetCursor(LoadCursor(hInst,"Hand"));
        }
        else
        {
            SetCursor(LoadCursor(0,MAKEINTRESOURCE(IDC_ARROW)));
        }
        return 0L;
    case WM_TIMER:
        if(moving)
        {
            GetCursorPos((LPPOINT)&lParam);
            ScreenToClient(hwnd,(LPPOINT)&lParam);
            GetClientRect(hwnd,&r);
            if((int)LOWORD(lParam)<=r.left)
            {
                Scroll(hwnd,SB_HORZ,SB_LINEUP,0);
            }
            else if((int)LOWORD(lParam)>r.right)
            {
                Scroll(hwnd,SB_HORZ,SB_LINEDOWN,0);
            }
            else if((int)HIWORD(lParam)<r.top)
            {
                Scroll(hwnd,SB_VERT,SB_LINEUP,0);
            }
            else if((int)HIWORD(lParam)>r.bottom)
            {
                Scroll(hwnd,SB_VERT,SB_LINEDOWN,0);
            }
        }
        return 0L;
    case WM_LBUTTONUP:
        if(timer)
        {
            KillTimer(hwnd,99);
        }
        if(moving>0)
        {
            Drop(hwnd,moving);
            moving=0;
        }
        ReleaseCapture(); /* Mauskontrolle abgeben */
        return 0L;
    case WM_PALETTECHANGED:
        if((HWND)wParam==hwnd) break;
        // fall thru
    case WM_QUERYNEWPALETTE:
        if(!palette) break;
        dc=GetDC(hwnd);
        oldpal=SelectPalette(dc,palette,FALSE);
        i=RealizePalette(dc);
        SelectPalette(dc,oldpal,FALSE);
        ReleaseDC(hwnd,dc);
        if(i) InvalidateRect(hwnd,NULL,TRUE);
        return i;
    case WM_PAINT:
        BeginPaint(hwnd,&ps);
        if(picture)
        {
            if(palette)
            {
                oldpal=SelectPalette(ps.hdc,palette,FALSE);
                RealizePalette(ps.hdc);
            }
            OffsetRect(&ps.rcPaint,-xoffset,-yoffset);
            Draw(ps.hdc,xoffset,yoffset,&ps.rcPaint);
            if(palette&&oldpal) SelectPalette(ps.hdc,oldpal,FALSE);
        }
        EndPaint(hwnd,&ps);
        return 0L;
    case WM_ERASEBKGND:
        hbr=CreateSolidBrush(color[backcolor]);
        GetClipBox((HDC)wParam,&r);
        FillRect((HDC)wParam,&r,hbr);
        DeleteObject(hbr);
        return TRUE;
    case WM_DESTROY:
        if(back) DeleteObject(back);
        WinHelp(hwnd,helpfile,HELP_QUIT,0L);
        if(shell>=32) FreeLibrary(shell);
        PostQuitMessage(0);
        return 0L;
#ifdef WIN32
    case WM_VSCROLL:
        Scroll(hwnd,SB_VERT,LOWORD(wParam),HIWORD(wParam));
        return 0L;
    case WM_HSCROLL:
        Scroll(hwnd,SB_HORZ,LOWORD(wParam),HIWORD(wParam));
        return 0L;
#else
    case WM_VSCROLL:
        Scroll(hwnd,SB_VERT,LOWORD(wParam),LOWORD(lParam));
        return 0L;
    case WM_HSCROLL:
        Scroll(hwnd,SB_HORZ,LOWORD(wParam),LOWORD(lParam));
        return 0L;
#endif
    }
    return DefWindowProc(hwnd,msg,wParam,lParam);
}

int PASCAL WinMain(HANDLE hThis,HANDLE hPrev,LPSTR lpCmdLine,int nCmdShow)
{
    WNDCLASS wc;
    MSG msg;
    OFSTRUCT of;
    int i,n;

    hInst=hThis;
    n=GetProfileString("Extensions","PUZ","",helpfile,sizeof(helpfile));
    i=GetModuleFileName(hInst,helpfile,sizeof(helpfile));
    if(n<1)
    {
        strcat(helpfile," ^.PUZ");
        WriteProfileString("Extensions","PUZ",helpfile);
    }
    while(i>0&&helpfile[i-1]!='.') i--;
    strcpy(helpfile+i,"HLP");
    if(!hPrev)
    {
        memset(&wc,0,sizeof(WNDCLASS));
        wc.lpfnWndProc=MainWndProc;
        wc.hInstance=hInst;
        wc.hIcon=LoadIcon(hInst,MAKEINTRESOURCE(1));
        if(GetProfileInt("intl","iCountry",0)==49)
        {
            addresid=46;
        }
        wc.lpszMenuName=MAKEINTRESOURCE(1+addresid);
        wc.lpszClassName="PuzzleWC";
        if(!RegisterClass(&wc)) return 0;
    }
    hFrame=CreateWindow("PuzzleWC",progname,WS_OVERLAPPEDWINDOW|WS_VSCROLL|WS_HSCROLL,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,0,0,hInst,NULL);
    if(hFrame)
    {
        masks[0][0][0]=LoadBitmap(hInst,"t0");
        masks[0][1][0]=LoadBitmap(hInst,"r0");
        masks[0][2][0]=LoadBitmap(hInst,"b0");
        masks[0][3][0]=LoadBitmap(hInst,"l0");
        masks[0][0][1]=LoadBitmap(hInst,"t1");
        masks[0][1][1]=LoadBitmap(hInst,"r2");
        masks[0][2][1]=LoadBitmap(hInst,"b2");
        masks[0][3][1]=LoadBitmap(hInst,"l1");
        masks[0][0][2]=LoadBitmap(hInst,"t2");
        masks[0][1][2]=LoadBitmap(hInst,"r1");
        masks[0][2][2]=LoadBitmap(hInst,"b1");
        masks[0][3][2]=LoadBitmap(hInst,"l2");
        masks[0][0][3]=LoadBitmap(hInst,"t3");
        masks[0][1][3]=LoadBitmap(hInst,"r4");
        masks[0][2][3]=LoadBitmap(hInst,"b4");
        masks[0][3][3]=LoadBitmap(hInst,"l3");
        masks[0][0][4]=LoadBitmap(hInst,"t4");
        masks[0][1][4]=LoadBitmap(hInst,"r3");
        masks[0][2][4]=LoadBitmap(hInst,"b3");
        masks[0][3][4]=LoadBitmap(hInst,"l4");
        masks[1][0][0]=LoadBitmap(hInst,"t048");
        masks[1][1][0]=LoadBitmap(hInst,"r048");
        masks[1][2][0]=LoadBitmap(hInst,"b048");
        masks[1][3][0]=LoadBitmap(hInst,"l048");
        masks[1][0][1]=LoadBitmap(hInst,"t148");
        masks[1][1][1]=LoadBitmap(hInst,"r148");
        masks[1][2][1]=LoadBitmap(hInst,"b148");
        masks[1][3][1]=LoadBitmap(hInst,"l148");
        masks[1][0][2]=LoadBitmap(hInst,"t248");
        masks[1][1][2]=LoadBitmap(hInst,"r248");
        masks[1][2][2]=LoadBitmap(hInst,"b248");
        masks[1][3][2]=LoadBitmap(hInst,"l248");
        masks[1][0][3]=LoadBitmap(hInst,"t348");
        masks[1][1][3]=LoadBitmap(hInst,"r348");
        masks[1][2][3]=LoadBitmap(hInst,"b348");
        masks[1][3][3]=LoadBitmap(hInst,"l348");
        masks[1][0][4]=LoadBitmap(hInst,"t448");
        masks[1][1][4]=LoadBitmap(hInst,"r448");
        masks[1][2][4]=LoadBitmap(hInst,"b448");
        masks[1][3][4]=LoadBitmap(hInst,"l448");
        hAccel=LoadAccelerators(hInst,MAKEINTRESOURCE(1));
        ShowWindow(hFrame,nCmdShow);
        UpdateWindow(hFrame);
        if(lpCmdLine||lpCmdLine[0]||GetProfileInt(progname,"HelloDone",FALSE)==0)
        {
            if(lpCmdLine||lpCmdLine[0])
            {
                i=OpenFile(lpCmdLine,&of,OF_EXIST);
            }
            else
            {
                i=OpenFile("PUZZLE.BMP",&of,OF_EXIST);
                hello=TRUE;
            }
            if(i!=-1)
            {
                LoadPicture(hFrame,of.szPathName);
            }
        }
        while(GetMessage(&msg,0,0,0))
        {
            if(!hAccel||!TranslateAccelerator(hFrame,hAccel,&msg))
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            }
        }
        FreePicture();
        for(n=0;n<4;n++) for(i=0;i<5;i++)
        {
            if(masks[0][n][i]) DeleteObject(masks[0][n][i]);
            if(masks[1][n][i]) DeleteObject(masks[1][n][i]);
        }
    }
    return 0;
}
