// Function RGBto256 converts true color bitmap to 256 color bitmap with
// optimal palette. Uses 512k histogram to build palette.

#include <stdlib.h>
#include <memory.h>
#include <windows.h>
#include "dib.h"

#ifndef abs
#define abs(x) ((x)<0?-(x):(x))
#endif

typedef unsigned int PANE[64];

static PANE far *hist[64];
static unsigned char rmin[256],rmax[256],gmin[256],gmax[256],bmin[256],bmax[256];
static unsigned char rmini[256],rmaxi[256],gmini[256],gmaxi[256],bmini[256],bmaxi[256];
static long count[256];
static RGBQUAD palette[256];

static BOOL nored(int r,int gmin,int gmax,int bmin,int bmax)
{
    int g,b;

    for(g=gmin;g<gmax;g++) for(b=bmin;b<bmax;b++)
    {
	if(hist[r][g][b]!=0) return FALSE;
    }
    return TRUE;
}

static BOOL nogreen(int rmin,int rmax,int g,int bmin,int bmax)
{
    int r,b;

    for(r=rmin;r<rmax;r++) for(b=bmin;b<bmax;b++)
    {
	if(hist[r][g][b]!=0) return FALSE;
    }
    return TRUE;
}

static BOOL noblue(int rmin,int rmax,int gmin,int gmax,int b)
{
    int r,g;

    for(r=rmin;r<rmax;r++) for(g=gmin;g<gmax;g++)
    {
	if(hist[r][g][b]!=0) return FALSE;
    }
    return TRUE;
}

static long countred(int r,int gmin,int gmax,int bmin,int bmax)
{
    int g,b;
    long sum;

    sum=0;
    for(g=gmin;g<gmax;g++) for(b=bmin;b<bmax;b++) sum+=hist[r][g][b];
    return sum;
}

static long countgreen(int rmin,int rmax,int g,int bmin,int bmax)
{
    int r,b;
    long sum;

    sum=0;
    for(r=rmin;r<rmax;r++) for(b=bmin;b<bmax;b++) sum+=hist[r][g][b];
    return sum;
}

static long countblue(int rmin,int rmax,int gmin,int gmax,int b)
{
    int r,g;
    long sum;

    sum=0;
    for(r=rmin;r<rmax;r++) for(g=gmin;g<gmax;g++) sum+=hist[r][g][b];
    return sum;
}

static void shrink(int i)
{
    while(rmin[i]<rmax[i]&&nored(rmin[i],gmin[i],gmax[i],bmin[i],bmax[i])) rmin[i]++;
    while(rmin[i]<rmax[i]&&nored(rmax[i]-1,gmin[i],gmax[i],bmin[i],bmax[i])) rmax[i]--;
    while(gmin[i]<gmax[i]&&nogreen(rmin[i],rmax[i],gmin[i],bmin[i],bmax[i])) gmin[i]++;
    while(gmin[i]<gmax[i]&&nogreen(rmin[i],rmax[i],gmax[i]-1,bmin[i],bmax[i])) gmax[i]--;
    while(bmin[i]<bmax[i]&&noblue(rmin[i],rmax[i],gmin[i],gmax[i],bmin[i])) bmin[i]++;
    while(bmin[i]<bmax[i]&&noblue(rmin[i],rmax[i],gmin[i],gmax[i],bmax[i]-1)) bmax[i]--;
}

static unsigned char lookup(int r,int g,int b)
{
    int i,n,d,m,k;
    RGBQUAD *p;

    n=hist[r>>2][g>>2][b>>2];
    if(n!=256) return n;
    n=0;
    p=&palette[n];
    d=r-p->rgbRed;
    m=abs(d);
    d=g-p->rgbGreen;
    d=abs(d);
    if(d>m) m=d;
    d=b-p->rgbBlue;
    d=abs(d);
    if(d>m) m=d;
    for(i=1;i<256;i++)
    {
	p++;
	d=r-p->rgbRed;
	k=abs(d);
	d=g-p->rgbGreen;
	d=abs(d);
	if(d>k) k=d;
	d=b-p->rgbBlue;
	d=abs(d);
	if(d>k) k=d;
	if(k<m)
	{
	    m=k;
	    n=i;
	}
    }
    hist[r>>2][g>>2][b>>2]=n;
    return n;
}

DIB *RGBto256(DIB *old)
{
    int x,y,i,n,j,width,height;
    long m,l,red,green,blue,pixel,skew;
    unsigned int far *h;
    unsigned char huge *scan;
    unsigned char huge *ptr;
    DIB *new;
    unsigned int size;
    unsigned char r,g,b,color;

    if(!old||!old->data||old->bmi.bmiHeader.biBitCount!=24) return NULL;
    size=sizeof(unsigned int)*64*64;
    for(i=0;i<64;i++) /* Hauptspeicher fr Histogramm */
    {
	hist[i]=(PANE far *)GlobalLock(GlobalAlloc(GMEM_MOVEABLE|GMEM_ZEROINIT,size));
	if(hist[i]==NULL)
	{
	    while(--i>=0)
	    {
	        GlobalUnlock(SEGMENT(hist[i]));
	        GlobalFree(SEGMENT(hist[i]));
	    }
	    return NULL;
	}
    }
    pixel=old->bmi.bmiHeader.biWidth*old->bmi.bmiHeader.biHeight;
    width=(int)old->bmi.bmiHeader.biWidth;
    height=(int)old->bmi.bmiHeader.biHeight;
    skew=((old->bmi.bmiHeader.biWidth*old->bmi.bmiHeader.biBitCount+31)/32)*4;
    for(y=0;y<height;y++)
    {
        scan=old->data+y*skew;
	for(x=0;x<width;x++)
	{
	    b=*scan++;
	    g=*scan++;
	    r=*scan++;
	    h=&hist[r>>2][g>>2][b>>2];
	    if(*h<65535U) (*h)++; else pixel--;
	}
    }
    rmin[0]=0; rmax[0]=64;
    gmin[0]=0; gmax[0]=64;
    bmin[0]=0; bmax[0]=64;
    rmini[0]=0; rmaxi[0]=64;
    gmini[0]=0; gmaxi[0]=64;
    bmini[0]=0; bmaxi[0]=64;
    count[0]=pixel;
    shrink(0);
    for(n=1;n<256;n++) /* Quader so oft unterteilen, bis gewnschte Farbanzahl */
    {
        i=0;
        m=rmax[0]-rmin[0];
        if(gmax[0]-gmin[0]>m) m=gmax[0]-gmin[0];
        if(bmax[0]-bmin[0]>m) m=bmax[0]-bmin[0];
        l=count[0]*m;
        for(j=1;j<256;j++)
        {
            m=rmax[j]-rmin[j];
            if(gmax[j]-gmin[j]>m) m=gmax[j]-gmin[j];
            if(bmax[j]-bmin[j]>m) m=bmax[j]-bmin[j];
            if(m>1&&count[j]*m>l) { l=count[j]*m; i=j; }
	}
	rmin[n]=rmin[i];
	rmax[n]=rmax[i];
	gmin[n]=gmin[i];
	gmax[n]=gmax[i];
	bmin[n]=bmin[i];
	bmax[n]=bmax[i];
	rmini[n]=rmini[i];
	rmaxi[n]=rmaxi[i];
	gmini[n]=gmini[i];
	gmaxi[n]=gmaxi[i];
	bmini[n]=bmini[i];
	bmaxi[n]=bmaxi[i];
	l=count[i]/2;
	if(rmax[i]-rmin[i]>gmax[i]-gmin[i]&&rmax[i]-rmin[i]>bmax[i]-bmin[i])
	{
	    for(pixel=countred(rmin[i]++,gmin[i],gmax[i],bmin[i],bmax[i]);rmin[i]<rmax[i]-1&&pixel<l;rmin[i]++)
	    {
		pixel+=countred(rmin[i],gmin[i],gmax[i],bmin[i],bmax[i]);
	    }
	    rmaxi[n]=rmax[n]=rmini[i]=rmin[i];
	}
	else if(gmax[i]-gmin[i]>bmax[i]-bmin[i])
	{
	    for(pixel=countgreen(rmin[i],rmax[i],gmin[i]++,bmin[i],bmax[i]);gmin[i]<gmax[i]-1&&pixel<l;gmin[i]++)
	    {
		pixel+=countgreen(rmin[i],rmax[i],gmin[i],bmin[i],bmax[i]);
	    }
	    gmaxi[n]=gmax[n]=gmini[i]=gmin[i];
	}
	else if(bmax[i]-bmin[i]>1)
	{
	    for(pixel=countblue(rmin[i],rmax[i],gmin[i],gmax[i],bmin[i]++);bmin[i]<bmax[i]-1&&pixel<l;bmin[i]++)
	    {
		pixel+=countblue(rmin[i],rmax[i],gmin[i],gmax[i],bmin[i]);
	    }
	    bmaxi[n]=bmax[n]=bmini[i]=bmin[i];
	}
	else break; /* keine weitere Farbaufteilung notwendig */
	count[i]-=pixel;
	count[n]=pixel;
	shrink(i);
	shrink(n);
    }
    for(i=0;i<n;i++)
    {
	if(count[i]>0L)
	{
	    red=green=blue=0L;
	    for(r=rmini[i];r<rmaxi[i];r++)
	    {
		for(g=gmini[i];g<gmaxi[i];g++)
		{
		    for(b=bmini[i];b<bmaxi[i];b++)
		    {
			l=hist[r][g][b]; /* long l weil Multiplikation long ergibt */
			if(l>0L)
			{
			    red+=l*r;
			    green+=l*g;
			    blue+=l*b;
			}
		    }
		}
	    }
	    pixel=count[i];
	    r=(unsigned char)(red/pixel);
	    g=(unsigned char)(green/pixel);
	    b=(unsigned char)(blue/pixel);
	    palette[i].rgbRed=r<<2;
	    palette[i].rgbGreen=g<<2;
	    palette[i].rgbBlue=b<<2;
	}
    }
    for(i=0;i<64;i++)
    {
	hist[i][0][0]=256;
	_fmemcpy(&hist[i][0][1],&hist[i][0][0],size-sizeof(int));
    }
    new=malloc(sizeof(DIB)+256*sizeof(RGBQUAD));
    if(new)
    {
        memcpy(new,old,sizeof(DIB));
        new->bmi.bmiHeader.biBitCount=8;
        new->bmi.bmiHeader.biClrUsed=256;
        new->bmi.bmiHeader.biClrImportant=n;
        l=((new->bmi.bmiHeader.biWidth*new->bmi.bmiHeader.biBitCount+31)/32)*4;
        new->bmi.bmiHeader.biSizeImage=l*new->bmi.bmiHeader.biHeight;
        if((new->data=GlobalLock(GlobalAlloc(GMEM_MOVEABLE,new->bmi.bmiHeader.biSizeImage)))!=NULL)
        {
            for(i=0;i<n;i++)
            {
                r=(new->bmi.bmiColors[i].rgbRed=palette[i].rgbRed)>>2;
                g=(new->bmi.bmiColors[i].rgbGreen=palette[i].rgbGreen)>>2;
                b=(new->bmi.bmiColors[i].rgbBlue=palette[i].rgbBlue)>>2;
                new->bmi.bmiColors[i].rgbReserved=0;
                hist[r][g][b]=i;
            }
            for(y=0;y<height;y++)
            {
                ptr=old->data+y*skew;
                scan=new->data+y*l;
                for(x=0;x<width;x++)
                {
		    b=*ptr++;
		    g=*ptr++;
		    r=*ptr++;
                    color=lookup(r,g,b);
                    *scan++=color;
                }
            }
	}
	else
	{
	    free(new);
	    new=NULL;
	}
    }
    for(i=0;i<64;i++)
    {
        GlobalUnlock(SEGMENT(hist[i]));
        GlobalFree(SEGMENT(hist[i]));
    }
    return new;
}
